@@ -1,5 +1,5 @@
-/* jdupes (C) 2015-2016 Jody Bruchon <jody@jodybruchon.com>
-   Derived from fdupes (C) 1999-2016 Adrian Lopez
+/* jdupes (C) 2015-2017 Jody Bruchon <jody@jodybruchon.com>
+   Derived from fdupes (C) 1999-2017 Adrian Lopez
 
    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation files
@@ -37,11 +37,13 @@
 #include <string.h>
 #include <errno.h>
 #include <libgen.h>
+#include <sys/time.h>
 #include "jdupes.h"
 #include "string_malloc.h"
 #include "jody_hash.h"
 #include "jody_sort.h"
 #include "jody_win_unicode.h"
+#include "jody_cacheinfo.h"
 #include "version.h"
 
 /* Headers for post-scanning actions */
@@ -83,51 +85,6 @@ int out_mode = _O_TEXT;
 #include "jody_paths.h"
 #endif
 
-#define ISFLAG(a,b) ((a & b) == b)
-#define SETFLAG(a,b) (a |= b)
-#define CLEARFLAG(a,b) (a &= (~b))
-
-/* Low memory option overrides */
-#ifdef LOW_MEMORY
- #undef DEBUG
- #undef LOUD_DEBUG
- #undef USE_TREE_REBALANCE
- #ifndef NO_PERMS
-  #define NO_PERMS 1
- #endif
-#endif
-
-/* Tree rebalance requires tree stats */
-#ifdef USE_TREE_REBALANCE
- #ifndef TREE_DEPTH_STATS
-  #define TREE_DEPTH_STATS
- #endif
-#endif
-
-/* Aggressive verbosity for deep debugging */
-#ifdef LOUD_DEBUG
- #ifndef DEBUG
-  #define DEBUG
- #endif
- #define LOUD(...) if ISFLAG(flags, F_LOUD) __VA_ARGS__
-#else
- #define LOUD(a)
-#endif
-
-/* Compile out debugging stat counters unless requested */
-#ifdef DEBUG
- #define DBG(a) a
- #ifndef TREE_DEPTH_STATS
-  #define TREE_DEPTH_STATS
- #endif
-#else
- #define DBG(a)
-#endif
-
-
-/* How many operations to wait before updating progress counters */
-#define DELAY_COUNT 256
-
 /* Behavior modification flags */
 uint_fast32_t flags = 0;
 
@@ -154,6 +111,8 @@ static enum {
  #define PARTIAL_HASH_SIZE 4096
 #endif
 
+static size_t auto_chunk_size;
+
 /* Maximum path buffer size to use; must be large enough for a path plus
  * any work that might be done to the array it's stored in. PATH_MAX is
  * not always true. Read this article on the false promises of PATH_MAX:
@@ -237,8 +196,11 @@ struct travdone {
 };
 static struct travdone *travdone_head = NULL;
 
-static uintmax_t filecount = 0; // Required for progress indicator code
-static int did_long_work = 0; // To tell progress indicator to go faster
+/* Required for progress indicator code */
+static uintmax_t filecount = 0;
+static uintmax_t progress = 0, dir_progress = 0, dupecount = 0;
+/* Number of read loops before checking progress indicator */
+#define CHECK_MINIMUM 256
 
 /* Hash/compare performance statistics (debug mode) */
 #ifdef DEBUG
@@ -273,6 +235,9 @@ static int sort_direction = 1;
 /* Signal handler */
 static int interrupt = 0;
 
+/* Progress indicator time */
+struct timeval time1, time2;
+
 
 /***** End definitions, begin code *****/
 
@@ -283,6 +248,7 @@ void sighandler(const int signum)
   (void)signum;
   if (interrupt || !ISFLAG(flags, F_SOFTABORT)) {
     fprintf(stderr, "\n");
+    string_malloc_destroy();
     exit(EXIT_FAILURE);
   }
   interrupt = 1;
@@ -294,10 +260,21 @@ void sighandler(const int signum)
 extern void oom(const char * const restrict msg)
 {
   fprintf(stderr, "\nout of memory: %s\n", msg);
+  string_malloc_destroy();
   exit(EXIT_FAILURE);
 }
 
 
+/* Null pointer failure */
+extern void nullptr(const char * restrict func)
+{
+  static const char n[] = "(NULL)";
+  if (func == NULL) func = n;
+  fprintf(stderr, "\ninternal error: NULL pointer passed to %s\n", func);
+  string_malloc_destroy();
+  exit(EXIT_FAILURE);
+}
+
 /* Compare two jody_hashes like memcmp() */
 #define HASH_COMPARE(a,b) ((a > b) ? 1:((a == b) ? 0:-1))
 
@@ -334,7 +311,7 @@ static int findarg(const char * const arg, const int start,
 
 /* Find the first non-option argument after specified option. */
 static int nonoptafter(const char *option, const int argc,
-                char **oldargv, char **newargv, int optind)
+                char **oldargv, char **newargv)
 {
   int x;
   int targetind;
@@ -353,10 +330,39 @@ static int nonoptafter(const char *option, const int argc,
 }
 
 
+/* Update progress indicator if requested */
+static void update_progress(const char * const restrict msg, const int file_percent)
+{
+  static int did_fpct = 0;
+
+  /* The caller should be doing this anyway...but don't trust that they did */
+  if (ISFLAG(flags, F_HIDEPROGRESS)) return;
+
+  gettimeofday(&time2, NULL);
+
+  if (progress == 0 || time2.tv_sec > time1.tv_sec) {
+    fprintf(stderr, "\rProgress [%ju/%ju, %ju pairs matched] %ju%%", progress, filecount,
+      dupecount, (progress * 100) / filecount);
+    if (file_percent > -1 && msg != NULL) {
+      fprintf(stderr, "  (%s: %d%%)         ", msg, file_percent);
+      did_fpct = 1;
+    } else if (did_fpct != 0) {
+      fprintf(stderr, "                     ");
+      did_fpct = 0;
+    }
+    fflush(stderr);
+  }
+  time1.tv_sec = time2.tv_sec;
+  return;
+}
+
 /* Check file's stat() info to make sure nothing has changed
  * Returns 1 if changed, 0 if not changed, negative if error */
 extern int file_has_changed(file_t * const restrict file)
 {
+  if (file == NULL || file->d_name == NULL) nullptr("file_has_changed()");
+  LOUD(fprintf(stderr, "file_has_changed('%s')\n", file->d_name);)
+
   if (!ISFLAG(file->flags, F_VALID_STAT)) return -66;
 
 #ifdef ON_WINDOWS
@@ -390,6 +396,9 @@ extern int file_has_changed(file_t * const restrict file)
 
 extern inline int getfilestats(file_t * const restrict file)
 {
+  if (file == NULL || file->d_name == NULL) nullptr("getfilestats()");
+  LOUD(fprintf(stderr, "getfilestats('%s')\n", file->d_name);)
+
   /* Don't stat the same file more than once */
   if (ISFLAG(file->flags, F_VALID_STAT)) return 0;
   SETFLAG(file->flags, F_VALID_STAT);
@@ -427,6 +436,9 @@ extern inline int getfilestats(file_t * const restrict file)
 extern int getdirstats(const char * const restrict name,
         jdupes_ino_t * const restrict inode, dev_t * const restrict dev)
 {
+  if (name == NULL || inode == NULL || dev == NULL) nullptr("getdirstats");
+  LOUD(fprintf(stderr, "getdirstats('%s', %p, %p)\n", name, (void *)inode, (void *)dev);)
+
 #ifdef ON_WINDOWS
   if (win_stat(name, &ws) != 0) return -1;
   *inode = ws.inode;
@@ -448,11 +460,9 @@ extern int getdirstats(const char * const restrict name,
  *  2 on an absolute match condition met */
 extern int check_conditions(const file_t * const restrict file1, const file_t * const restrict file2)
 {
-  /* Null pointer sanity check */
-  if (!file1 || !file2) {
-    LOUD(fprintf(stderr, "internal error: null pointer passed to check_conditions()\n");)
-    exit(EXIT_FAILURE);
-  }
+  if (file1 == NULL || file2 == NULL || file1->d_name == NULL || file2->d_name == NULL) nullptr("check_conditions()");
+
+  LOUD(fprintf(stderr, "check_conditions('%s', '%s')\n", file1->d_name, file2->d_name);)
 
   /* Exclude based on -I/--isolate */
   if (ISFLAG(flags, F_ISOLATE) && (file1->user_order == file2->user_order)) {
@@ -487,11 +497,11 @@ extern int check_conditions(const file_t * const restrict file1, const file_t *
 
   /* Exclude files that are not the same size */
   if (file1->size > file2->size) {
-    LOUD(fprintf(stderr, "check_conditions: no match: file1 > file2 (%jd < %jd)\n", (intmax_t)file1->size, (intmax_t)file2->size));
+    LOUD(fprintf(stderr, "check_conditions: no match: size of file1 > file2 (%jd > %jd)\n", (intmax_t)file1->size, (intmax_t)file2->size));
     return -1;
   }
   if (file1->size < file2->size) {
-    LOUD(fprintf(stderr, "check_conditions: no match: file1 < file2 (%jd > %jd)\n", (intmax_t)file1->size, (intmax_t)file2->size));
+    LOUD(fprintf(stderr, "check_conditions: no match: size of file1 < file2 (%jd < %jd)\n", (intmax_t)file1->size, (intmax_t)file2->size));
     return 1;
   }
 
@@ -501,18 +511,24 @@ extern int check_conditions(const file_t * const restrict file1, const file_t *
 }
 
 
-/* Create a new traversal check object */
-static int travdone_alloc(struct travdone **trav, const char * const restrict dir)
+/* Create a new traversal check object and initialize its values */
+static struct travdone *travdone_alloc(const jdupes_ino_t inode, const dev_t device)
 {
-  *trav = string_malloc(sizeof(struct travdone));
-  if (*trav == NULL) {
-    LOUD(fprintf(stderr, "travdone_alloc: malloc returned NULL\n");)
-    return -1;
+  struct travdone *trav;
+
+  LOUD(fprintf(stderr, "travdone_alloc(%jd, %jd)\n", (intmax_t)inode, (intmax_t)device);)
+
+  trav = (struct travdone *)string_malloc(sizeof(struct travdone));
+  if (trav == NULL) {
+    LOUD(fprintf(stderr, "travdone_alloc: malloc failed\n");)
+    return NULL;
   }
-  (*trav)->left = NULL;
-  (*trav)->right = NULL;
-  if (getdirstats(dir, &((*trav)->inode), &((*trav)->device)) != 0) return -1;
-  return 0;
+  trav->left = NULL;
+  trav->right = NULL;
+  trav->inode = inode;
+  trav->device = device;
+  LOUD(fprintf(stderr, "travdone_alloc returned %p\n", (void *)trav);)
+  return trav;
 }
 
 
@@ -526,9 +542,7 @@ static void grokdir(const char * const restrict dir,
   static struct stat linfo;
 #endif
   struct dirent *dirinfo;
-  static uintmax_t progress = 0, dir_progress = 0;
   static int grokdir_level = 0;
-  static int delay = DELAY_COUNT;
   static char tempname[PATHBUF_SIZE * 2];
   size_t dirlen;
   struct travdone *traverse;
@@ -542,25 +556,28 @@ static void grokdir(const char * const restrict dir,
   DIR *cd;
 #endif
 
+  if (dir == NULL || filelistp == NULL) nullptr("grokdir()");
   LOUD(fprintf(stderr, "grokdir: scanning '%s' (order %d)\n", dir, user_dir_count));
 
   /* Double traversal prevention tree */
+  if (getdirstats(dir, &inode, &device) != 0) goto error_travdone;
   if (travdone_head == NULL) {
-    if (travdone_alloc(&travdone_head, dir) != 0) goto error_travdone;
+    travdone_head = travdone_alloc(inode, device);
+    if (travdone_head == NULL) goto error_travdone;
   } else {
-    if (getdirstats(dir, &inode, &device) != 0) goto error_travdone;
     traverse = travdone_head;
     while (1) {
+      if (traverse == NULL) nullptr("grokdir() traverse");
       /* Don't re-traverse directories we've already seen */
       if (inode == traverse->inode && device == traverse->device) {
         LOUD(fprintf(stderr, "already seen dir '%s', skipping\n", dir);)
         return;
-      }
-      if (inode >= traverse->inode) {
+      } else if (inode > traverse->inode || (inode == traverse->inode && device > traverse->device)) {
         /* Traverse right */
         if (traverse->right == NULL) {
           LOUD(fprintf(stderr, "traverse dir right '%s'\n", dir);)
-          if (travdone_alloc(&(traverse->right), dir) != 0) goto error_travdone;
+          traverse->right = travdone_alloc(inode, device);
+          if (traverse->right == NULL) goto error_travdone;
           break;
         }
         traverse = traverse->right;
@@ -569,7 +586,8 @@ static void grokdir(const char * const restrict dir,
         /* Traverse left */
         if (traverse->left == NULL) {
           LOUD(fprintf(stderr, "traverse dir left '%s'\n", dir);)
-          if (travdone_alloc(&(traverse->left), dir) != 0) goto error_travdone;
+          traverse->left = travdone_alloc(inode, device);
+          if (traverse->left == NULL) goto error_travdone;
           break;
         }
         traverse = traverse->left;
@@ -593,7 +611,7 @@ static void grokdir(const char * const restrict dir,
 
   LOUD(fprintf(stderr, "FindFirstFile: %s\n", dir));
   hFind = FindFirstFile((LPCWSTR)wname, &ffd);
-  if (hFind == INVALID_HANDLE_VALUE) { fprintf(stderr, "handle bad\n"); goto error_cd; }
+  if (hFind == INVALID_HANDLE_VALUE) { fprintf(stderr, "\nfile handle bad\n"); goto error_cd; }
   LOUD(fprintf(stderr, "Loop start\n"));
   do {
     char * restrict tp = tempname;
@@ -614,11 +632,12 @@ static void grokdir(const char * const restrict dir,
     LOUD(fprintf(stderr, "grokdir: readdir: '%s'\n", dirinfo->d_name));
     if (strcmp(dirinfo->d_name, ".") && strcmp(dirinfo->d_name, "..")) {
       if (!ISFLAG(flags, F_HIDEPROGRESS)) {
-        if (delay >= DELAY_COUNT) {
-          delay = 0;
+        gettimeofday(&time2, NULL);
+        if (progress == 0 || time2.tv_sec > time1.tv_sec) {
           fprintf(stderr, "\rScanning: %ju files, %ju dirs (in %u specified)",
               progress, dir_progress, user_dir_count);
-        } else delay++;
+        }
+        time1.tv_sec = time2.tv_sec;
       }
 
       /* Assemble the file's full path name, optimized to avoid strcat() */
@@ -777,16 +796,16 @@ static void grokdir(const char * const restrict dir,
 
   grokdir_level--;
   if (grokdir_level == 0 && !ISFLAG(flags, F_HIDEPROGRESS)) {
-    fprintf(stderr, "\rExamining %ju files, %ju dirs (in %u specified)",
+    fprintf(stderr, "\rScanning: %ju files, %ju dirs (in %u specified)",
             progress, dir_progress, user_dir_count);
   }
   return;
 
 error_travdone:
-  fprintf(stderr, "error: could not stat dir "); fwprint(stderr, dir, 1);
+  fprintf(stderr, "\ncould not stat dir "); fwprint(stderr, dir, 1);
   return;
 error_cd:
-  fprintf(stderr, "error: could not chdir to "); fwprint(stderr, dir, 1);
+  fprintf(stderr, "\ncould not chdir to "); fwprint(stderr, dir, 1);
   return;
 error_overflow:
   fprintf(stderr, "\nerror: a path buffer overflowed\n");
@@ -802,6 +821,10 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
   static hash_t hash[1];
   static hash_t chunk[(CHUNK_SIZE / sizeof(hash_t))];
   FILE *file;
+  int check = 0;
+
+  if (checkfile == NULL || checkfile->d_name == NULL) nullptr("get_filehash()");
+  LOUD(fprintf(stderr, "get_filehash('%s', %jd)\n", checkfile->d_name, (intmax_t)max_read);)
 
   /* Get the file size. If we can't read it, bail out early */
   if (checkfile->size == -1) {
@@ -838,7 +861,7 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
   file = fopen(checkfile->d_name, FILE_MODE_RO);
 #endif
   if (file == NULL) {
-    fprintf(stderr, "error opening file "); fwprint(stderr, checkfile->d_name, 1);
+    fprintf(stderr, "\nerror opening file "); fwprint(stderr, checkfile->d_name, 1);
     return NULL;
   }
   /* Actually seek past the first chunk if applicable
@@ -846,7 +869,7 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
   if (ISFLAG(checkfile->flags, F_HASH_PARTIAL)) {
     if (fseeko(file, PARTIAL_HASH_SIZE, SEEK_SET) == -1) {
       fclose(file);
-      fprintf(stderr, "error seeking in file "); fwprint(stderr, checkfile->d_name, 1);
+      fprintf(stderr, "\nerror seeking in file "); fwprint(stderr, checkfile->d_name, 1);
       return NULL;
     }
     fsize -= PARTIAL_HASH_SIZE;
@@ -856,9 +879,9 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
     size_t bytes_to_read;
 
     if (interrupt) return 0;
-    bytes_to_read = (fsize >= CHUNK_SIZE) ? CHUNK_SIZE : (size_t)fsize;
+    bytes_to_read = (fsize >= (off_t)auto_chunk_size) ? auto_chunk_size : (size_t)fsize;
     if (fread((void *)chunk, bytes_to_read, 1, file) != 1) {
-      fprintf(stderr, "error reading from file "); fwprint(stderr, checkfile->d_name, 1);
+      fprintf(stderr, "\nerror reading from file "); fwprint(stderr, checkfile->d_name, 1);
       fclose(file);
       return NULL;
     }
@@ -866,6 +889,14 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
     *hash = jody_block_hash(chunk, *hash, bytes_to_read);
     if ((off_t)bytes_to_read > fsize) break;
     else fsize -= (off_t)bytes_to_read;
+
+    if (!ISFLAG(flags, F_HIDEPROGRESS)) {
+      check++;
+      if (check > CHECK_MINIMUM) {
+        update_progress("hashing", (int)(((checkfile->size - fsize) * 100) / checkfile->size));
+        check = 0;
+      }
+    }
   }
 
   fclose(file);
@@ -880,6 +911,9 @@ static inline void registerfile(filetree_t * restrict * const restrict nodeptr,
 {
   filetree_t * restrict branch;
 
+  if (nodeptr == NULL || file == NULL || (d != NONE && *nodeptr == NULL)) nullptr("registerfile()");
+  LOUD(fprintf(stderr, "registerfile(direction %d)\n", d));
+
   /* Allocate and initialize a new node for the file */
   branch = (filetree_t *)string_malloc(sizeof(filetree_t));
   if (branch == NULL) oom("registerfile() branch");
@@ -908,6 +942,10 @@ static inline void registerfile(filetree_t * restrict * const restrict nodeptr,
       *nodeptr = branch;
       break;
     default:
+      /* This should never ever happen */
+      fprintf(stderr, "\ninternal error: invalid direction for registerfile(), report this\n");
+      string_malloc_destroy();
+      exit(EXIT_FAILURE);
       break;
   }
 
@@ -919,13 +957,13 @@ static inline void registerfile(filetree_t * restrict * const restrict nodeptr,
     if (up->left == branch) up->left_weight++;
     else if (up->right == branch) up->right_weight++;
     else {
-      fprintf(stderr, "Internal error: file tree linkage is broken\n");
+      fprintf(stderr, "\nInternal error: file tree linkage is broken\n");
       exit(EXIT_FAILURE);
     }
     branch = up;
   }
 #else /* USE_TREE_REBALANCE */
-  /* Attach the new node to the requested branch and the parent */
+  /* Attach the new node to the requested branch */
   switch (d) {
     case LEFT:
       (*nodeptr)->left = branch;
@@ -934,10 +972,15 @@ static inline void registerfile(filetree_t * restrict * const restrict nodeptr,
       (*nodeptr)->right = branch;
       break;
     case NONE:
-    default:
       /* For the root of the tree only */
       *nodeptr = branch;
       break;
+    default:
+      /* This should never ever happen */
+      fprintf(stderr, "\ninternal error: invalid direction for registerfile(), report this\n");
+      string_malloc_destroy();
+      exit(EXIT_FAILURE);
+      break;
   }
 
 #endif /* USE_TREE_REBALANCE */
@@ -1056,7 +1099,8 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
   int cmpresult = 0;
   const hash_t * restrict filehash;
 
-  LOUD(fprintf(stderr, "checkmatch (\"%s\", \"%s\")\n", tree->file->d_name, file->d_name));
+  if (tree == NULL || file == NULL || tree->file == NULL || tree->file->d_name == NULL || file->d_name == NULL) nullptr("checkmatch()");
+  LOUD(fprintf(stderr, "checkmatch ('%s', '%s')\n", tree->file->d_name, file->d_name));
 
   /* If device and inode fields are equal one of the files is a
    * hard link to the other or the files have been listed twice
@@ -1084,10 +1128,7 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
     /* Attempt to exclude files quickly with partial file hashing */
     if (!ISFLAG(tree->file->flags, F_HASH_PARTIAL)) {
       filehash = get_filehash(tree->file, PARTIAL_HASH_SIZE);
-      if (filehash == NULL) {
-        if (!interrupt) fprintf(stderr, "cannot read file "); fwprint(stderr, tree->file->d_name, 1);
-        return NULL;
-      }
+      if (filehash == NULL) return NULL;
 
       tree->file->filehash_partial = *filehash;
       SETFLAG(tree->file->flags, F_HASH_PARTIAL);
@@ -1095,10 +1136,7 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
 
     if (!ISFLAG(file->flags, F_HASH_PARTIAL)) {
       filehash = get_filehash(file, PARTIAL_HASH_SIZE);
-      if (filehash == NULL) {
-        if (!interrupt) fprintf(stderr, "cannot read file "); fwprint(stderr, file->d_name, 1);
-        return NULL;
-      }
+      if (filehash == NULL) return NULL;
 
       file->filehash_partial = *filehash;
       SETFLAG(file->flags, F_HASH_PARTIAL);
@@ -1125,24 +1163,16 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
     } else if (cmpresult == 0) {
       /* If partial match was correct, perform a full file hash match */
       if (!ISFLAG(tree->file->flags, F_HASH_FULL)) {
-        did_long_work = 1;
         filehash = get_filehash(tree->file, 0);
-        if (filehash == NULL) {
-          if (!interrupt) fprintf(stderr, "cannot read file "); fwprint(stderr, tree->file->d_name, 1);
-          return NULL;
-        }
+        if (filehash == NULL) return NULL;
 
         tree->file->filehash = *filehash;
         SETFLAG(tree->file->flags, F_HASH_FULL);
       }
 
       if (!ISFLAG(file->flags, F_HASH_FULL)) {
-        did_long_work = 1;
         filehash = get_filehash(file, 0);
-        if (filehash == NULL) {
-          if (!interrupt) fprintf(stderr, "cannot read file "); fwprint(stderr, file->d_name, 1);
-          return NULL;
-        }
+        if (filehash == NULL) return NULL;
 
         file->filehash = *filehash;
         SETFLAG(file->flags, F_HASH_FULL);
@@ -1194,26 +1224,35 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
 
 /* Do a byte-by-byte comparison in case two different files produce the
    same signature. Unlikely, but better safe than sorry. */
-static inline int confirmmatch(FILE * const restrict file1, FILE * const restrict file2)
+static inline int confirmmatch(FILE * const restrict file1, FILE * const restrict file2, off_t size)
 {
-  static char c1[CHUNK_SIZE];
-  static char c2[CHUNK_SIZE];
-  size_t r1;
-  size_t r2;
+  static char c1[CHUNK_SIZE], c2[CHUNK_SIZE];
+  size_t r1, r2;
+  off_t bytes = 0;
+  int check = 0;
 
+  if (file1 == NULL || file2 == NULL) nullptr("confirmmatch()");
   LOUD(fprintf(stderr, "confirmmatch running\n"));
 
-  did_long_work = 1;
   fseek(file1, 0, SEEK_SET);
   fseek(file2, 0, SEEK_SET);
 
   do {
     if (interrupt) return 0;
-    r1 = fread(c1, sizeof(char), CHUNK_SIZE, file1);
-    r2 = fread(c2, sizeof(char), CHUNK_SIZE, file2);
+    r1 = fread(c1, sizeof(char), auto_chunk_size, file1);
+    r2 = fread(c2, sizeof(char), auto_chunk_size, file2);
 
     if (r1 != r2) return 0; /* file lengths are different */
     if (memcmp (c1, c2, r1)) return 0; /* file contents are different */
+
+    if (!ISFLAG(flags, F_HIDEPROGRESS)) {
+      check++;
+      bytes += (off_t)r1;
+      if (check > CHECK_MINIMUM) {
+        update_progress("confirm", (int)((bytes * 100) / size));
+        check = 0;
+      }
+    }
   } while (r2);
 
   return 1;
@@ -1228,6 +1267,9 @@ extern unsigned int get_max_dupes(const file_t *files, unsigned int * const rest
                 unsigned int * const restrict n_files) {
   unsigned int groups = 0;
 
+  if (files == NULL || max == NULL) nullptr("get_max_dupes()");
+  LOUD(fprintf(stderr, "get_max_dupes(%p, %p, %p)\n", (const void *)files, (void *)max, (void *)n_files));
+
   *max = 0;
   if (n_files) *n_files = 0;
 
@@ -1249,6 +1291,7 @@ extern unsigned int get_max_dupes(const file_t *files, unsigned int * const rest
 static int sort_pairs_by_param_order(file_t *f1, file_t *f2)
 {
   if (!ISFLAG(flags, F_USEPARAMORDER)) return 0;
+  if (f1 == NULL || f2 == NULL) nullptr("sort_pairs_by_param_order()");
   if (f1->user_order < f2->user_order) return -sort_direction;
   if (f1->user_order > f2->user_order) return sort_direction;
   return 0;
@@ -1257,6 +1300,7 @@ static int sort_pairs_by_param_order(file_t *f1, file_t *f2)
 
 static int sort_pairs_by_mtime(file_t *f1, file_t *f2)
 {
+  if (f1 == NULL || f2 == NULL) nullptr("sort_pairs_by_mtime()");
   int po = sort_pairs_by_param_order(f1, f2);
 
   if (po != 0) return po;
@@ -1270,6 +1314,7 @@ static int sort_pairs_by_mtime(file_t *f1, file_t *f2)
 
 static int sort_pairs_by_filename(file_t *f1, file_t *f2)
 {
+  if (f1 == NULL || f2 == NULL) nullptr("sort_pairs_by_filename()");
   int po = sort_pairs_by_param_order(f1, f2);
 
   if (po != 0) return po;
@@ -1285,9 +1330,9 @@ static void registerpair(file_t **matchlist, file_t *newmatch,
   file_t *back;
 
   /* NULL pointer sanity checks */
-  if (!matchlist || !newmatch || !comparef) return;
-
+  if (matchlist == NULL || newmatch == NULL || comparef == NULL) nullptr("registerpair()");
   LOUD(fprintf(stderr, "registerpair: '%s', '%s'\n", (*matchlist)->d_name, newmatch->d_name);)
+
   SETFLAG((*matchlist)->flags, F_HAS_DUPES);
   back = NULL;
   traverse = *matchlist;
@@ -1398,6 +1443,7 @@ int wmain(int argc, wchar_t **wargv)
 int main(int argc, char **argv)
 #endif
 {
+  static struct proc_cacheinfo pci;
   static file_t *files = NULL;
   static file_t *curfile;
   static char **oldargv;
@@ -1454,6 +1500,12 @@ int main(int argc, char **argv)
 #define GETOPT getopt
 #endif
 
+/* Windows buffers our stderr output; don't let it do that */
+#ifdef ON_WINDOWS
+  if (setvbuf(stderr, NULL, _IONBF, 0) != 0)
+    fprintf(stderr, "warning: setvbuf() failed\n");
+#endif
+
 #ifdef UNICODE
   /* Create a UTF-8 **argv from the wide version */
   static char **argv;
@@ -1465,6 +1517,16 @@ int main(int argc, char **argv)
   else out_mode = _O_U16TEXT;
 #endif /* UNICODE */
 
+  /* Auto-tune chunk size to be half of L1 data cache if possible */
+  get_proc_cacheinfo(&pci);
+  if (pci.l1 != 0) auto_chunk_size = (pci.l1 / 2);
+  else if (pci.l1d != 0) auto_chunk_size = (pci.l1d / 2);
+  /* Must be at least 4096 (4 KiB) and cannot exceed CHUNK_SIZE */
+  if (auto_chunk_size < 4096 || auto_chunk_size > CHUNK_SIZE) auto_chunk_size = CHUNK_SIZE;
+  /* Force to a multiple of 4096 if it isn't already */
+  if ((auto_chunk_size & 0x00000fffUL) != 0)
+    auto_chunk_size = (auto_chunk_size + 0x00000fffUL) & 0x000ff000;
+
   program_name = argv[0];
 
   oldargv = cloneargs(argc, argv);
@@ -1492,6 +1554,7 @@ int main(int argc, char **argv)
       break;
     case 'h':
       help_text();
+      string_malloc_destroy();
       exit(EXIT_FAILURE);
 #ifndef NO_HARDLINKS
     case 'H':
@@ -1588,7 +1651,17 @@ int main(int argc, char **argv)
 #endif
       break;
     case 'v':
-      printf("jdupes %s (%s)\n", VER, VERDATE);
+      printf("jdupes %s (%s) ", VER, VERDATE);
+
+      /* Indicate bitness information */
+      if (sizeof(uintptr_t) == 8) {
+        if (sizeof(long) == 4) printf("64-bit i32\n");
+        else if (sizeof(long) == 8) printf("64-bit\n");
+      } else if (sizeof(uintptr_t) == 4) {
+        if (sizeof(long) == 4) printf("32-bit\n");
+        else if (sizeof(long) == 8) printf("32-bit i64\n");
+      } else printf("%ld-bit i%ld\n", sizeof(uintptr_t) * 8, sizeof(long) * 8);
+
       printf("Compile-time extensions:");
       if (*extensions != NULL) {
         int c = 0;
@@ -1597,7 +1670,7 @@ int main(int argc, char **argv)
           c++;
         }
       } else printf(" none");
-      printf("\nCopyright (C) 2015-2016 by Jody Bruchon\n");
+      printf("\nCopyright (C) 2015-2017 by Jody Bruchon\n");
       printf("\nPermission is hereby granted, free of charge, to any person\n");
       printf("obtaining a copy of this software and associated documentation files\n");
       printf("(the \"Software\"), to deal in the Software without restriction,\n");
@@ -1640,27 +1713,32 @@ int main(int argc, char **argv)
 
     default:
       fprintf(stderr, "Try `jdupes --help' for more information.\n");
+      string_malloc_destroy();
       exit(EXIT_FAILURE);
     }
   }
 
   if (optind >= argc) {
     fprintf(stderr, "no directories specified (use -h option for help)\n");
+    string_malloc_destroy();
     exit(EXIT_FAILURE);
   }
 
   if (ISFLAG(flags, F_ISOLATE) && optind == (argc - 1)) {
     fprintf(stderr, "Isolation requires at least two directories on the command line\n");
+    string_malloc_destroy();
     exit(EXIT_FAILURE);
   }
 
   if (ISFLAG(flags, F_RECURSE) && ISFLAG(flags, F_RECURSEAFTER)) {
     fprintf(stderr, "options --recurse and --recurse: are not compatible\n");
+    string_malloc_destroy();
     exit(EXIT_FAILURE);
   }
 
   if (ISFLAG(flags, F_SUMMARIZEMATCHES) && ISFLAG(flags, F_DELETEFILES)) {
     fprintf(stderr, "options --summarize and --delete are not compatible\n");
+    string_malloc_destroy();
     exit(EXIT_FAILURE);
   }
 
@@ -1678,18 +1756,20 @@ int main(int argc, char **argv)
 
   if (pm > 1) {
       fprintf(stderr, "Only one of --summarize, --delete, --linkhard, --linksoft, or --dedupe\nmay be used\n");
+      string_malloc_destroy();
       exit(EXIT_FAILURE);
   }
   if (pm == 0) SETFLAG(flags, F_PRINTMATCHES);
 
   if (ISFLAG(flags, F_RECURSEAFTER)) {
-    firstrecurse = nonoptafter("--recurse:", argc, oldargv, argv, optind);
+    firstrecurse = nonoptafter("--recurse:", argc, oldargv, argv);
 
     if (firstrecurse == argc)
-      firstrecurse = nonoptafter("-R", argc, oldargv, argv, optind);
+      firstrecurse = nonoptafter("-R", argc, oldargv, argv);
 
     if (firstrecurse == argc) {
       fprintf(stderr, "-R option must be isolated from other options\n");
+      string_malloc_destroy();
       exit(EXIT_FAILURE);
     }
 
@@ -1721,17 +1801,15 @@ int main(int argc, char **argv)
   if (!files) exit(EXIT_SUCCESS);
 
   curfile = files;
+  progress = 0;
 
   /* Catch CTRL-C */
   signal(SIGINT, sighandler);
 
   while (curfile) {
-    static uintmax_t progress = 0;
-    static uintmax_t dupecount = 0;
     static file_t **match = NULL;
     static FILE *file1;
     static FILE *file2;
-    static off_t delay = DELAY_COUNT;
 #ifdef USE_TREE_REBALANCE
     static unsigned int depth_threshold = INITIAL_DEPTH_THRESHOLD;
 #endif
@@ -1798,7 +1876,7 @@ int main(int argc, char **argv)
         continue;
       }
 
-      if (confirmmatch(file1, file2)) {
+      if (confirmmatch(file1, file2, curfile->size)) {
         LOUD(fprintf(stderr, "MAIN: registering matched file pair\n"));
         registerpair(match, curfile,
             (ordertype == ORDER_TIME) ? sort_pairs_by_mtime : sort_pairs_by_filename);
@@ -1812,20 +1890,7 @@ int main(int argc, char **argv)
 skip_full_check:
     curfile = curfile->next;
 
-    if (!ISFLAG(flags, F_HIDEPROGRESS)) {
-      /* If file size is larger than 1 MiB, make progress update faster
-       * If confirmmatch() is run on a file, speed up progress even further */
-      if (curfile != NULL && did_long_work) {
-        delay += (curfile->size >> 20);
-        did_long_work = 0;
-      }
-      if (match != NULL) delay++;
-      if ((delay >= DELAY_COUNT)) {
-        delay = 0;
-        fprintf(stderr, "\rProgress [%ju/%ju, %ju pairs matched] %ju%%", progress, filecount,
-          dupecount, (progress * 100) / filecount);
-      } else delay++;
-    }
+    if (!ISFLAG(flags, F_HIDEPROGRESS)) update_progress(NULL, -1);
     progress++;
   }
 
@@ -1863,6 +1928,8 @@ skip_file_scan:
     fprintf(stderr, "Max tree depth: %u; SMA: allocs %ju, free %ju, fail %ju, reuse %ju, scan %ju, tails %ju\n",
         max_depth, sma_allocs, sma_free_good, sma_free_ignored,
         sma_free_reclaimed, sma_free_scanned, sma_free_tails);
+    fprintf(stderr, "I/O chunk size: %ju KiB (%s)\n", (uintmax_t)(auto_chunk_size >> 10),
+        (pci.l1 + pci.l1d) != 0 ? "dynamically sized" : "default size");
 #ifdef ON_WINDOWS
  #ifndef NO_HARDLINKS
     if (ISFLAG(flags, F_HARDLINKFILES))
