@@ -1,4 +1,5 @@
-/* jdupes module header */
+/* jdupes main program header
+ * See jdupes.c for license information */
 
 #ifndef JDUPES_H
 #define JDUPES_H
@@ -77,8 +78,6 @@ extern int out_mode;
 
 /* Low memory option overrides */
 #ifdef LOW_MEMORY
- #undef DEBUG
- #undef LOUD_DEBUG
  #undef USE_TREE_REBALANCE
  #ifndef NO_PERMS
   #define NO_PERMS 1
@@ -150,10 +149,6 @@ typedef enum {
   ORDER_TIME
 } ordertype_t;
 
-/* Larger chunk size makes large files process faster but uses more RAM */
-#ifndef CHUNK_SIZE
- #define CHUNK_SIZE 65536
-#endif
 #ifndef PARTIAL_HASH_SIZE
  #define PARTIAL_HASH_SIZE 4096
 #endif
@@ -215,6 +210,7 @@ extern struct stat s;
 #endif
 
 extern void oom(const char * const restrict msg);
+extern void nullptr(const char * restrict func);
 extern int file_has_changed(file_t * const restrict file);
 extern int getfilestats(file_t * const restrict file);
 extern int getdirstats(const char * const restrict name,
